@@ -1,4 +1,4 @@
-// Copyright (c) 2020-2022, Sylabs Inc. All rights reserved.
+// Copyright (c) 2020-2024, Sylabs Inc. All rights reserved.
 // This software is licensed under a 3-clause BSD license. Please consult the LICENSE.md file
 // distributed with the sources of this project regarding your rights to use or distribute this
 // software.
@@ -6,6 +6,7 @@
 package integrity
 
 import (
+	"context"
 	"crypto"
 	"errors"
 	"io"
@@ -21,7 +22,7 @@ import (
 
 func TestGroupVerifier_signatures(t *testing.T) {
 	oneGroupImage := loadContainer(t, filepath.Join(corpus, "one-group.sif"))
-	oneGroupSignedImage := loadContainer(t, filepath.Join(corpus, "one-group-signed.sif"))
+	oneGroupSignedImage := loadContainer(t, filepath.Join(corpus, "one-group-signed-pgp.sif"))
 
 	sigs, err := oneGroupSignedImage.GetDescriptors(sif.WithDataType(sif.DataSignature))
 	if err != nil {
@@ -50,7 +51,6 @@ func TestGroupVerifier_signatures(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			v := &groupVerifier{
 				f:       tt.f,
@@ -71,7 +71,7 @@ func TestGroupVerifier_signatures(t *testing.T) {
 }
 
 func TestGroupVerifier_verify(t *testing.T) {
-	oneGroupSignedImage := loadContainer(t, filepath.Join(corpus, "one-group-signed.sif"))
+	oneGroupSignedImage := loadContainer(t, filepath.Join(corpus, "one-group-signed-pgp.sif"))
 
 	sig, err := oneGroupSignedImage.GetDescriptor(sif.WithDataType(sif.DataSignature))
 	if err != nil {
@@ -143,7 +143,6 @@ func TestGroupVerifier_verify(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			ods := make([]sif.Descriptor, len(tt.objectIDs))
 			for i, id := range tt.objectIDs {
@@ -162,7 +161,7 @@ func TestGroupVerifier_verify(t *testing.T) {
 			}
 
 			var vr VerifyResult
-			err := v.verifySignature(tt.sig, tt.de, &vr)
+			err := v.verifySignature(context.Background(), tt.sig, tt.de, &vr)
 
 			if got, want := err, tt.wantErr; !errors.Is(got, want) {
 				t.Errorf("got error %v, want %v", got, want)
@@ -210,7 +209,6 @@ func TestLegacyGroupVerifier_signatures(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			v := &legacyGroupVerifier{
 				f:       tt.f,
@@ -278,7 +276,6 @@ func TestLegacyGroupVerifier_verify(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			ods, err := getGroupObjects(tt.f, tt.groupID)
 			if err != nil {
@@ -292,7 +289,7 @@ func TestLegacyGroupVerifier_verify(t *testing.T) {
 			}
 
 			var vr VerifyResult
-			err = v.verifySignature(tt.sig, tt.de, &vr)
+			err = v.verifySignature(context.Background(), tt.sig, tt.de, &vr)
 
 			if got, want := err, tt.wantErr; !errors.Is(got, want) {
 				t.Errorf("got error %v, want %v", got, want)
@@ -343,7 +340,6 @@ func TestLegacyObjectVerifier_signatures(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			od, err := tt.f.GetDescriptor(sif.WithID(tt.id))
 			if err != nil {
@@ -419,7 +415,6 @@ func TestLegacyObjectVerifier_verify(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			od, err := tt.f.GetDescriptor(sif.WithID(tt.id))
 			if err != nil {
@@ -432,7 +427,7 @@ func TestLegacyObjectVerifier_verify(t *testing.T) {
 			}
 
 			var vr VerifyResult
-			err = v.verifySignature(tt.sig, tt.de, &vr)
+			err = v.verifySignature(context.Background(), tt.sig, tt.de, &vr)
 
 			if got, want := err, tt.wantErr; !errors.Is(got, want) {
 				t.Errorf("got error %v, want %v", got, want)
@@ -449,21 +444,24 @@ func TestLegacyObjectVerifier_verify(t *testing.T) {
 	}
 }
 
-func TestNewVerifier(t *testing.T) {
+func TestNewVerifier(t *testing.T) { //nolint:maintidx
 	emptyImage := loadContainer(t, filepath.Join(corpus, "empty.sif"))
 	oneGroupImage := loadContainer(t, filepath.Join(corpus, "one-group.sif"))
 	twoGroupImage := loadContainer(t, filepath.Join(corpus, "two-groups.sif"))
 
+	sv := getTestVerifier(t, "ed25519-public.pem", crypto.Hash(0))
+
 	kr := openpgp.EntityList{getTestEntity(t)}
 
-	cb := func(r VerifyResult) bool { return false }
+	cb := func(_ VerifyResult) bool { return false }
 
 	tests := []struct {
 		name          string
 		fi            *sif.FileImage
 		opts          []VerifierOpt
 		wantErr       error
-		wantKeyring   openpgp.KeyRing
+		wantDSSE      decoder
+		wantClearsign decoder
 		wantGroups    []uint32
 		wantObjects   []uint32
 		wantLegacy    bool
@@ -539,12 +537,32 @@ func TestNewVerifier(t *testing.T) {
 			wantTasks:  2,
 		},
 		{
-			name:        "OptVerifyWithKeyRing",
-			fi:          twoGroupImage,
-			opts:        []VerifierOpt{OptVerifyWithKeyRing(kr)},
-			wantKeyring: kr,
-			wantGroups:  []uint32{1, 2},
-			wantTasks:   2,
+			name:       "OptVerifyWithVerifier",
+			fi:         twoGroupImage,
+			opts:       []VerifierOpt{OptVerifyWithVerifier(sv)},
+			wantDSSE:   newDSSEDecoder(sv),
+			wantGroups: []uint32{1, 2},
+			wantTasks:  2,
+		},
+		{
+			name:          "OptVerifyWithKeyRing",
+			fi:            twoGroupImage,
+			opts:          []VerifierOpt{OptVerifyWithKeyRing(kr)},
+			wantClearsign: newClearsignDecoder(kr),
+			wantGroups:    []uint32{1, 2},
+			wantTasks:     2,
+		},
+		{
+			name: "OptVerifyWithVerifierAndKeyRing",
+			fi:   twoGroupImage,
+			opts: []VerifierOpt{
+				OptVerifyWithVerifier(sv),
+				OptVerifyWithKeyRing(kr),
+			},
+			wantDSSE:      newDSSEDecoder(sv),
+			wantClearsign: newClearsignDecoder(kr),
+			wantGroups:    []uint32{1, 2},
+			wantTasks:     2,
 		},
 		{
 			name:       "OptVerifyGroupDuplicate",
@@ -653,7 +671,6 @@ func TestNewVerifier(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			v, err := NewVerifier(tt.fi, tt.opts...)
 			if got, want := err, tt.wantErr; !errors.Is(got, want) {
@@ -665,8 +682,12 @@ func TestNewVerifier(t *testing.T) {
 					t.Errorf("got FileImage %v, want %v", got, want)
 				}
 
-				if got, want := v.kr, tt.wantKeyring; !reflect.DeepEqual(got, want) {
-					t.Errorf("got key ring %v, want %v", got, want)
+				if got, want := v.dsse, tt.wantDSSE; !reflect.DeepEqual(got, want) {
+					t.Errorf("got DSSE decoder %+v, want %+v", got, want)
+				}
+
+				if got, want := v.cs, tt.wantClearsign; !reflect.DeepEqual(got, want) {
+					t.Errorf("got clear-sign decoder %+v, want %+v", got, want)
 				}
 
 				if got, want := len(v.tasks), tt.wantTasks; got != want {
@@ -690,7 +711,7 @@ func (v mockVerifier) signatures() ([]sif.Descriptor, error) {
 	return v.sigs, v.sigsErr
 }
 
-func (v mockVerifier) verifySignature(sig sif.Descriptor, de decoder, vr *VerifyResult) error {
+func (v mockVerifier) verifySignature(_ context.Context, _ sif.Descriptor, _ decoder, vr *VerifyResult) error {
 	vr.verified = v.verified
 	vr.e = v.e
 	return v.verifyErr
@@ -817,7 +838,6 @@ func TestVerifier_AnySignedBy(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			v := Verifier{tasks: tt.tasks}
 
@@ -913,7 +933,6 @@ func TestVerifier_AllSignedBy(t *testing.T) {
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
 			v := Verifier{tasks: tt.tasks}
 
@@ -932,18 +951,31 @@ func TestVerifier_AllSignedBy(t *testing.T) {
 
 func TestVerifier_Verify(t *testing.T) {
 	oneGroupImage := loadContainer(t, filepath.Join(corpus, "one-group.sif"))
-	oneGroupSignedImage := loadContainer(t, filepath.Join(corpus, "one-group-signed.sif"))
+	oneGroupSignedPGPImage := loadContainer(t, filepath.Join(corpus, "one-group-signed-pgp.sif"))
+	oneGroupSignedDSSEImage := loadContainer(t, filepath.Join(corpus, "one-group-signed-dsse.sif"))
 
-	verified, err := oneGroupSignedImage.GetDescriptors(sif.WithGroupID(1))
+	verifiedDSSE, err := oneGroupSignedDSSEImage.GetDescriptors(sif.WithGroupID(1))
 	if err != nil {
 		t.Fatal(err)
 	}
 
-	sig, err := oneGroupSignedImage.GetDescriptor(sif.WithDataType(sif.DataSignature))
+	sigDSSE, err := oneGroupSignedDSSEImage.GetDescriptor(sif.WithDataType(sif.DataSignature))
 	if err != nil {
 		t.Fatal(err)
 	}
 
+	verifiedPGP, err := oneGroupSignedPGPImage.GetDescriptors(sif.WithGroupID(1))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	sigPGP, err := oneGroupSignedPGPImage.GetDescriptor(sif.WithDataType(sif.DataSignature))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	ed25519 := getTestVerifier(t, "ed25519-public.pem", crypto.Hash(0))
+
 	e := getTestEntity(t)
 
 	kr := openpgp.EntityList{e}
@@ -951,127 +983,152 @@ func TestVerifier_Verify(t *testing.T) {
 	tests := []struct {
 		name            string
 		f               *sif.FileImage
-		tasks           []verifyTask
-		kr              openpgp.KeyRing
+		opts            []VerifierOpt
 		testCallback    bool
 		ignoreError     bool
 		wantCBSignature sif.Descriptor
 		wantCBVerified  []sif.Descriptor
+		wantCBKeys      []crypto.PublicKey
 		wantCBEntity    *openpgp.Entity
 		wantCBErr       error
 		wantErr         error
 	}{
 		{
-			name: "NoKeyMaterial",
-			f:    oneGroupSignedImage,
-			tasks: []verifyTask{
-				mockVerifier{},
-			},
-			wantErr: ErrNoKeyMaterial,
-		},
-		{
-			name: "SignatureNotFound",
-			f:    oneGroupImage,
-			tasks: []verifyTask{
-				mockVerifier{
-					sigsErr: &SignatureNotFoundError{ID: 1, IsGroup: true},
-				},
-			},
-			kr:      kr,
+			name:    "SignatureNotFound",
+			f:       oneGroupImage,
 			wantErr: &SignatureNotFoundError{},
 		},
 		{
-			name: "UnknownIssuer",
-			f:    oneGroupSignedImage,
-			tasks: []verifyTask{
-				mockVerifier{
-					sigs:      []sif.Descriptor{sig},
-					verifyErr: &SignatureNotValidError{ID: 3, Err: pgperrors.ErrUnknownIssuer},
-				},
+			name: "NoKeyMaterialDSSE",
+			f:    oneGroupSignedDSSEImage,
+			opts: []VerifierOpt{
+				OptVerifyWithKeyRing(kr),
 			},
-			kr:      kr,
-			wantErr: &SignatureNotValidError{},
+			wantErr: errNoKeyMaterialDSSE,
 		},
 		{
-			name: "UnknownIssuerIgnoreError",
-			f:    oneGroupSignedImage,
-			tasks: []verifyTask{
-				mockVerifier{
-					sigs:      []sif.Descriptor{sig},
-					verifyErr: &SignatureNotValidError{ID: 3, Err: pgperrors.ErrUnknownIssuer},
-				},
+			name: "NoKeyMaterialPGP",
+			f:    oneGroupSignedPGPImage,
+			opts: []VerifierOpt{
+				OptVerifyWithVerifier(ed25519),
+			},
+			wantErr: errNoKeyMaterialPGP,
+		},
+		{
+			name: "SignatureNotValidErrorDSSE",
+			f:    oneGroupSignedDSSEImage,
+			opts: []VerifierOpt{
+				OptVerifyWithVerifier(
+					getTestVerifier(t, "ecdsa-public.pem", crypto.SHA256), // Not signed with ECDSA.
+				),
+			},
+			wantErr: &SignatureNotValidError{ID: 3},
+		},
+		{
+			name: "SignatureNotValidErrorPGP",
+			f:    oneGroupSignedPGPImage,
+			opts: []VerifierOpt{
+				OptVerifyWithKeyRing(openpgp.EntityList{}),
+			},
+			wantErr: &SignatureNotValidError{ID: 3},
+		},
+		{
+			name: "OneGroupSignedDSSE",
+			f:    oneGroupSignedDSSEImage,
+			opts: []VerifierOpt{
+				OptVerifyWithVerifier(ed25519),
+			},
+		},
+		{
+			name: "OneGroupSignedPGP",
+			f:    oneGroupSignedPGPImage,
+			opts: []VerifierOpt{
+				OptVerifyWithKeyRing(kr),
+			},
+		},
+		{
+			name: "OneGroupSignedDSSEWithCallback",
+			f:    oneGroupSignedDSSEImage,
+			opts: []VerifierOpt{
+				OptVerifyWithVerifier(ed25519),
+			},
+			testCallback:    true,
+			wantCBSignature: sigDSSE,
+			wantCBVerified:  verifiedDSSE,
+			wantCBKeys: []crypto.PublicKey{
+				getTestPublicKey(t, "ed25519-public.pem"),
+			},
+			wantCBEntity: nil,
+		},
+		{
+			name: "OneGroupSignedPGPWithCallback",
+			f:    oneGroupSignedPGPImage,
+			opts: []VerifierOpt{
+				OptVerifyWithKeyRing(kr),
+			},
+			testCallback:    true,
+			wantCBSignature: sigPGP,
+			wantCBVerified:  verifiedPGP,
+			wantCBKeys:      []crypto.PublicKey{},
+			wantCBEntity:    e,
+		},
+		{
+			name: "OneGroupSignedPGPWithCallbackIgnoreError",
+			f:    oneGroupSignedPGPImage,
+			opts: []VerifierOpt{
+				OptVerifyWithKeyRing(openpgp.EntityList{}),
 			},
 			testCallback:    true,
 			ignoreError:     true,
-			kr:              openpgp.EntityList{},
-			wantCBSignature: sig,
-			wantCBErr:       &SignatureNotValidError{ID: 3, Err: pgperrors.ErrUnknownIssuer},
-			wantErr:         nil,
-		},
-		{
-			name: "OneGroupSigned",
-			f:    oneGroupSignedImage,
-			tasks: []verifyTask{
-				mockVerifier{
-					sigs:     []sif.Descriptor{sig},
-					verified: verified,
-				},
-			},
-			kr: kr,
-		},
-		{
-			name:         "OneGroupSignedWithCallback",
-			f:            oneGroupSignedImage,
-			testCallback: true,
-			tasks: []verifyTask{
-				mockVerifier{
-					sigs:     []sif.Descriptor{sig},
-					verified: verified,
-					e:        e,
-				},
-			},
-			kr:              kr,
-			wantCBSignature: sig,
-			wantCBVerified:  verified,
-			wantCBEntity:    e,
+			wantCBSignature: sigPGP,
+			wantCBKeys:      []crypto.PublicKey{},
+			wantCBEntity:    nil,
+			wantCBErr:       &SignatureNotValidError{ID: 3},
 		},
 	}
 
 	for _, tt := range tests {
-		tt := tt
 		t.Run(tt.name, func(t *testing.T) {
-			v := Verifier{
-				f:     tt.f,
-				tasks: tt.tasks,
-				kr:    tt.kr,
+			var vr VerifyResult
+
+			opts := tt.opts
+			if tt.testCallback {
+				opts = append(opts, OptVerifyCallback(func(r VerifyResult) bool {
+					vr = r
+					return tt.ignoreError
+				}))
 			}
 
-			// Test callback functionality, if requested.
-			if tt.testCallback {
-				v.cb = func(r VerifyResult) bool {
-					if got, want := r.Signature(), tt.wantCBSignature; got != want {
-						t.Errorf("got signature %v, want %v", got, want)
-					}
-
-					if got, want := r.Verified(), tt.wantCBVerified; !reflect.DeepEqual(got, want) {
-						t.Errorf("got verified %v, want %v", got, want)
-					}
-
-					if got, want := r.Entity(), tt.wantCBEntity; got != want {
-						t.Errorf("got entity %v, want %v", got, want)
-					}
-
-					if got, want := r.Error(), tt.wantCBErr; !errors.Is(got, want) {
-						t.Errorf("got error %v, want %v", got, want)
-					}
-
-					return tt.ignoreError
-				}
+			v, err := NewVerifier(tt.f, opts...)
+			if err != nil {
+				t.Fatal(err)
 			}
 
 			if got, want := v.Verify(), tt.wantErr; !errors.Is(got, want) {
 				t.Fatalf("got error %v, want %v", got, want)
 			}
+
+			if tt.testCallback {
+				if got, want := vr.Signature(), tt.wantCBSignature; got != want {
+					t.Errorf("got signature %v, want %v", got, want)
+				}
+
+				if got, want := vr.Verified(), tt.wantCBVerified; !reflect.DeepEqual(got, want) {
+					t.Errorf("got verified %v, want %v", got, want)
+				}
+
+				if got, want := vr.Keys(), tt.wantCBKeys; !reflect.DeepEqual(got, want) {
+					t.Errorf("got keys %v, want %v", got, want)
+				}
+
+				if got, want := vr.Entity(), tt.wantCBEntity; got != want {
+					t.Errorf("got entity %v, want %v", got, want)
+				}
+
+				if got, want := vr.Error(), tt.wantCBErr; !errors.Is(got, want) {
+					t.Errorf("got error %v, want %v", got, want)
+				}
+			}
 		})
 	}
 }
